#include "stdio.h"
#include "conio.h"
#include "stdlib.h"
#include "windows.h"
#include "math.h"
#include "assert.h"

#ifndef EPS
#define EPS 0.0000001
#endif

struct Pnt
{
	double x;
	double y;
};

class Matrix {

	public:

	// Default Constructor. Creates a 1 by 1 matrix; sets value to zero. 
	Matrix () {
	nRow_ = 1; nCol_ = 1;
	data_ = new double [1];  // Allocate memory
	set(0.0);                // Set value of data_[0] to 0.0
	}

	// Regular Constructor. Creates an nR by nC matrix; sets values to zero.
	// If number of columns is not specified, it is set to 1.
	Matrix(int nR, int nC = 1) {
	assert(nR > 0 && nC > 0);    // Check that nC and nR both > 0.
	nRow_ = nR; nCol_ = nC;
	data_ = new double [nR*nC];  // Allocate memory
	assert(data_ != 0);          // Check that memory was allocated
	set(0.0);                    // Set values of data_[] to 0.0
	}

	// Copy Constructor.
	// Used when a copy of an object is produced 
	// (e.g., passing to a function by value)
	Matrix(const Matrix& mat) {
	this->copy(mat);   // Call private copy function.
	}

	// Destructor. Called when a Matrix object goes out of scope.
	~Matrix() {
	delete [] data_;   // Release allocated memory
	}

	// Assignment operator function.
	// Overloads the equal sign operator to work with
	// Matrix objects.
	Matrix& operator=(const Matrix& mat) {
	if( this == &mat ) return *this;  // If two sides equal, do nothing.
	delete [] data_;                  // Delete data on left hand side
	this->copy(mat);                  // Copy right hand side to l.h.s.
	return *this;
	}

	// Simple "get" functions. Return number of rows or columns.
	int nRow() const { return nRow_; }
	int nCol() const { return nCol_; }

	// Parenthesis operator function.
	// Allows access to values of Matrix via (i,j) pair.
	// Example: a(1,1) = 2*b(2,3); 
	// If column is unspecified, take as 1.
	double& operator() (int i, int j = 1) {
	assert(i > 0 && i <= nRow_);          // Bounds checking for rows
	assert(j > 0 && j <= nCol_);          // Bounds checking for columns
	return data_[ nCol_*(i-1) + (j-1) ];  // Access appropriate value
	}

	// Parenthesis operator function (const version).
	const double& operator() (int i, int j = 1) const{
	assert(i > 0 && i <= nRow_);          // Bounds checking for rows
	assert(j > 0 && j <= nCol_);          // Bounds checking for columns
	return data_[ nCol_*(i-1) + (j-1) ];  // Access appropriate value
	}

	// Set function. Sets all elements of a matrix to a given value.
	void set(double value) {
	int i, iData = nRow_*nCol_;
	for( i=0; i<iData; i++ )
		data_[i] = value;
	}

	//*********************************************************************
	private:

	// Matrix data.
	int nRow_, nCol_;  // Number of rows, columns
	double* data_;     // Pointer used to allocate memory for data.

	// Private copy function.
	// Copies values from one Matrix object to another.
	void copy(const Matrix& mat) {
	nRow_ = mat.nRow_;
	nCol_ = mat.nCol_;
	int i, iData = nRow_*nCol_;
	data_ = new double [iData];
	for(i = 0; i<iData; i++ )
		data_[i] = mat.data_[i];
	}

}; // Class Matrix


// Compute inverse of matrix
double inv(Matrix A, Matrix& Ainv) 
// Input
//    A    -    Matrix A (N by N)
// Outputs
//   Ainv  -    Inverse of matrix A (N by N)
//  determ -    Determinant of matrix A	(return value)
{

  int N = A.nRow();
  assert( N == A.nCol() );
  
  Ainv = A;  // Copy matrix to ensure Ainv is same size
    
  int i, j, k;
  Matrix scale(N), b(N,N);	 // Scale factor and work array
  int *index;  index = new int [N+1];

  //* Matrix b is initialized to the identity matrix
  b.set(0.0);
  for( i=1; i<=N; i++ )
    b(i,i) = 1.0;

  //* Set scale factor, scale(i) = max( |a(i,j)| ), for each row
  for( i=1; i<=N; i++ ) {
    index[i] = i;			  // Initialize row index list
    double scalemax = 0.;
    for( j=1; j<=N; j++ ) 
      scalemax = (scalemax > fabs(A(i,j))) ? scalemax : fabs(A(i,j));
    scale(i) = scalemax;
  }

  //* Loop over rows k = 1, ..., (N-1)
  int signDet = 1;
  for( k=1; k<=N-1; k++ ) {
	//* Select pivot row from max( |a(j,k)/s(j)| )
    double ratiomax = 0.0;
	int jPivot = k;
    for( i=k; i<=N; i++ ) {
      double ratio = fabs(A(index[i],k))/scale(index[i]);
      if( ratio > ratiomax ) {
        jPivot=i;
        ratiomax = ratio;
      }
    }
	//* Perform pivoting using row index list
	int indexJ = index[k];
	if( jPivot != k ) {	          // Pivot
      indexJ = index[jPivot];
      index[jPivot] = index[k];   // Swap index jPivot and k
      index[k] = indexJ;
	  signDet *= -1;			  // Flip sign of determinant
	}
	//* Perform forward elimination
    for( i=k+1; i<=N; i++ ) {
      double coeff = A(index[i],k)/A(indexJ,k);
      for( j=k+1; j<=N; j++ )
        A(index[i],j) -= coeff*A(indexJ,j);
      A(index[i],k) = coeff;
      for( j=1; j<=N; j++ ) 
        b(index[i],j) -= A(index[i],k)*b(indexJ,j);
    }
  }
  //* Compute determinant as product of diagonal elements
  double determ = signDet;	   // Sign of determinant
  for( i=1; i<=N; i++ )
	determ *= A(index[i],i);

  //* Perform backsubstitution
  for( k=1; k<=N; k++ ) {
    Ainv(N,k) = b(index[N],k)/A(index[N],N);
    for( i=N-1; i>=1; i--) {
      double sum = b(index[i],k);
      for( j=i+1; j<=N; j++ )
        sum -= A(index[i],j)*Ainv(j,k);
      Ainv(i,k) = sum/A(index[i],i);
    }
  }

  delete [] index;	// Release allocated memory
  return( determ );        
}

struct Vector3
{
	double x;
	double y;
	double z;
};

struct Vector3 Cross(struct Vector3 v1, struct Vector3 v2)
{
	struct Vector3 v3;
	v3.x = v1.y*v2.z - v1.z*v2.y;
	v3.y = v1.z*v2.x - v1.x*v2.z;
	v3.z = v1.x*v2.y - v1.y*v2.x;
	return v3;
}

struct Vector3 Normalize(struct Vector3 v)
{
	Vector3 _v;
	_v.x = v.x/v.z;
	_v.y = v.y/v.z;
	_v.z = 1;
	return _v;
}

void main()
{
	long i,j;
	// Load the image
	FILE *fp;
	fp = fopen("5.ppm","rb");
	if(fp==NULL)
	{
		printf("cannot open the 1.ppm\n");
		return;
	}

	char pszbuff[1024];
	int width,height,color;
	fgets(pszbuff,1024,fp);
	fscanf(fp,"%d %d\n",&width,&height);
	fscanf(fp,"%d\n",&color);

	BYTE *pR,*pG,*pB;
	long size = width*height;
	pR = new BYTE[size];
	pG = new BYTE[size];
	pB = new BYTE[size];
	
	for(i=0;i<size;i++)
	{
		fread(&pR[i],sizeof(BYTE),1,fp);
		fread(&pG[i],sizeof(BYTE),1,fp);
		fread(&pB[i],sizeof(BYTE),1,fp);
	}

	fclose(fp);

	// Calculate the matrix of transformation
	struct Vector3 ProjV[4];
	struct Vector3 IntersectV1,IntersectV2;
	struct Vector3 InfLine;
	double l1,l2,l3;

	ProjV[0].x = 103; ProjV[0].y = 69;  ProjV[0].z = 1;
	ProjV[1].x = 360; ProjV[1].y = 118; ProjV[1].z = 1;
	ProjV[2].x = 373; ProjV[2].y = 329; ProjV[2].z = 1;
	ProjV[3].x = 107; ProjV[3].y = 357; ProjV[3].z = 1;

	struct Vector3 line1,line2;
	line1 = Cross(ProjV[0],ProjV[3]);
	line2 = Cross(ProjV[1],ProjV[2]);
	
	IntersectV1 = Cross(Cross(ProjV[0],ProjV[3]),Cross(ProjV[1],ProjV[2]));
	IntersectV2 = Cross(Cross(ProjV[0],ProjV[1]),Cross(ProjV[2],ProjV[3]));

	InfLine = Normalize(Cross(IntersectV1,IntersectV2));
	l1 = InfLine.x;
	l2 = InfLine.y;
	l3 = InfLine.z;
	/***********************************
		|		1			0		0|
	H = |		0			1		0| 
		|-(l1/l3)	 -(l2/l3)  	 1/l3|
	************************************/

	BYTE *pTR,*pTG,*pTB;
	pTR = new BYTE[size];
	pTG = new BYTE[size];
	pTB = new BYTE[size];

	memset(pTR,0,size);
	memset(pTG,0,size);
	memset(pTB,0,size);

	for(i=0;i<height;i++)
	{
		for(j=0;j<width;j++)
		{
			double x,y,_x,_y;
			x = (double)(j*2.13);
			y = (double)(i*2.13);
			_x = x/(-x*l1/l3-y*l2/l3+1/l3);
			_y = y/(-x*l1/l3-y*l2/l3+1/l3);
			if(_x<1 || _x>=(width-1) || _y<1 || _y>=(height-1))
				continue;

			long Oindex = (long)(i*width+j);

			// interpolation
			double fx,fy;
			fx = floor(_x);
			fy = floor(_y);
			double r11,r12,r21,r22;
			r11 = (fx+1-_x)*(fy+1-_y);
			r12 = (_x-fx)*(fy+1-_y);
			r21 = (fx+1-_x)*(_y-fy);
			r22 = (_x-fx)*(_y-fy);
			
			pTR[Oindex] = (BYTE)(pR[(long)fy*width+(long)fx]*r11+pR[(long)fy*width+(long)(fx+1)]*r12+pR[(long)(fy+1)*width+(long)fx]*r21+pR[(long)(fy+1)*width+(long)(fx+1)]*r22);
			pTG[Oindex] = (BYTE)(pG[(long)fy*width+(long)fx]*r11+pG[(long)fy*width+(long)(fx+1)]*r12+pG[(long)(fy+1)*width+(long)fx]*r21+pG[(long)(fy+1)*width+(long)(fx+1)]*r22);
			pTB[Oindex] = (BYTE)(pB[(long)fy*width+(long)fx]*r11+pB[(long)fy*width+(long)(fx+1)]*r12+pB[(long)(fy+1)*width+(long)fx]*r21+pB[(long)(fy+1)*width+(long)(fx+1)]*r22);
		}
	}

	fclose(fp);

/*
	double maxx,minx;
	double maxy,miny;

	maxx = maxy = -9999999.0;
	minx = miny =  9999999.0;

	for(i=0;i<height;i++)
	{
		for(j=0;j<width;j++)
		{
			BYTE r=255,g=255,b=255;
			double x,y,_x,_y;
			x = (double)j;
			y = (double)i;
			_x = x/(x*l1+y*l2+l3);
			_y = y/(x*l1+y*l2+l3);

			if(_x>maxx) maxx = _x;
			if(_x<minx) minx = _x;
			if(_y>maxy) maxy = _y;
			if(_y<miny) miny = _y;


		}
	}

	fp = fopen("out.txt","w");

	fprintf(fp,"%12.9f\n",(float)(maxx));
	fprintf(fp,"%12.9f\n",(float)(minx));
	fprintf(fp,"%12.9f\n",(float)(maxy));
	fprintf(fp,"%12.9f\n",(float)(miny));

	fclose(fp);
*/

	// write the image
	fp = fopen("5-1.ppm","wb");
	char str[3][40];
	strcpy(str[0],"P6");
	sprintf(str[1],"%d %d",width,height);
	strcpy(str[2],"255");
	BYTE ret = 10;
	for(int t=0;t<3;t++)
	{
		int len = strlen(str[t]);
		for(i=0;i<len;i++)
			fwrite(&str[t][i],1,1,fp);
		fwrite(&ret,1,1,fp);
	}
	for(i=0;i<size;i++)
	{
		fwrite(&pTR[i],1,1,fp);
		fwrite(&pTG[i],1,1,fp);
		fwrite(&pTB[i],1,1,fp);
	}
	fclose(fp);

	return;
}